@@ -1,3 +1,5 @@
+use std::collections::HashSet;
+
 use proc_macro2::{Span, TokenStream};
 use quote::{quote, quote_spanned};
 use syn::spanned::Spanned;
@@ -6,6 +8,20 @@ use crate::internals::attr::{self, DefaultTag, FieldAttr, Packing, TypeAttr};
 use crate::internals::symbol::*;
 use crate::internals::{Ctxt, Needs, NeedsKind};
 
+#[derive(Clone, Copy, PartialEq, Eq, Hash)]
+enum TagMethod {
+    /// The default tag method.
+    Default,
+    /// Special method that requires generating a visitor.
+    String,
+}
+
+impl Default for TagMethod {
+    fn default() -> Self {
+        Self::Default
+    }
+}
+
 struct Tokens {
     decode_t: TokenStream,
     decoder_t: TokenStream,
@@ -20,7 +36,9 @@ struct Tokens {
     pair_decoder_t: TokenStream,
     pair_encoder_t: TokenStream,
     struct_decoder_t: TokenStream,
-    variant_encoder_t: TokenStream,
+    reference_visitor_t: TokenStream,
+    phantom_data: TokenStream,
+    fmt: TokenStream,
 }
 
 pub(crate) struct Expander<'a> {
@@ -55,8 +73,10 @@ impl<'a> Expander<'a> {
                 pack_encoder_t: quote!(#prefix::en::PackEncoder),
                 pair_decoder_t: quote!(#prefix::de::PairDecoder),
                 pair_encoder_t: quote!(#prefix::en::PairEncoder),
-                struct_decoder_t: quote!(#prefix::de::StructDecoder),
-                variant_encoder_t: quote!(#prefix::en::VariantEncoder),
+                struct_decoder_t: quote!(#prefix::de::PairsDecoder),
+                reference_visitor_t: quote!(#prefix::de::ValueVisitor),
+                phantom_data: quote!(core::marker::PhantomData),
+                fmt: quote!(core::fmt),
             },
         }
     }
@@ -109,7 +129,7 @@ impl<'a> Expander<'a> {
             #[automatically_derived]
             impl #type_generics #encode_t for #type_ident #type_generics {
                 #inline
-                fn encode<E>(&self, #assignment: E) -> Result<(), E::Error>
+                fn encode<E>(&self, #assignment: E) -> Result<E::Ok, E::Error>
                 where
                     E: #encoder_t
                 {
@@ -173,8 +193,7 @@ impl<'a> Expander<'a> {
         let (span, encode_path) = field_attr.encode_path(encode_t, span);
 
         Some(quote_spanned! {
-            span =>
-            #encode_path(#accessor, #encoder_var)?;
+            span => #encode_path(#accessor, #encoder_var)
         })
     }
 
@@ -212,8 +231,7 @@ impl<'a> Expander<'a> {
         let (span, encode_path) = field_attr.encode_path(encode_t, span);
 
         let encode = quote_spanned! {
-            span =>
-            #encode_path(this, #encoder_var)?;
+            span => #encode_path(this, #encoder_var)
         };
 
         Some((encode, vec![pattern]))
@@ -306,15 +324,8 @@ impl<'a> Expander<'a> {
             }
             Packing::Tagged => {
                 needs.mark_used();
-
-                let (setup, finish) = self.encode_field_tag(fields, &field_tests)?;
-
-                quote! {
-                    #(#decls)*
-                    let mut #encoder_var = #setup;
-                    #(#encoders)*
-                    #finish
-                }
+                let encode = quote! { #(#encoders)* };
+                self.encode_field_tag(fields, &decls, encode, &field_tests)
             }
             Packing::Packed => {
                 needs.mark_used();
@@ -326,15 +337,12 @@ impl<'a> Expander<'a> {
                     #(#decls)*
                     let mut pack = #encoder_t::encode_pack(#encoder_var)?;
                     #(#encoders)*
-                    #pack_encoder_t::finish(pack)?;
+                    #pack_encoder_t::end(pack)
                 }
             }
         };
 
-        Some(quote! {
-            #encode
-            Ok(())
-        })
+        Some(encode)
     }
 
     fn encode_enum(&self, data: &syn::DataEnum, needs: &mut Needs) -> Option<TokenStream> {
@@ -364,8 +372,6 @@ impl<'a> Expander<'a> {
                 match self {
                     #(#variants)*
                 }
-
-                Ok(())
             }
         })
     }
@@ -580,7 +586,7 @@ impl<'a> Expander<'a> {
 
             variants.push(quote! {
                 #tag => {
-                    let #decoder_var = #pair_decoder_t::decode_second(variant)?;
+                    let #decoder_var = #pair_decoder_t::second(variant)?;
                     #output
                 }
             });
@@ -598,14 +604,14 @@ impl<'a> Expander<'a> {
             Some(ident) => {
                 quote! {
                     if !#pair_decoder_t::skip_second(variant)? {
-                        return Err(<D::Error as #error_t>::unsupported_variant(#type_name, tag));
+                        return Err(<D::Error as #error_t>::invalid_variant_tag(#type_name, tag));
                     }
 
                     Self::#ident {}
                 }
             }
             None => quote! {
-                return Err(<D::Error as #error_t>::unsupported_variant(#type_name, tag));
+                return Err(<D::Error as #error_t>::invalid_variant_tag(#type_name, tag));
             },
         };
 
@@ -613,7 +619,7 @@ impl<'a> Expander<'a> {
             let mut variant = #decoder_t::decode_variant(#decoder_var)?;
 
             let #variant_tag #tag_type = {
-                let tag_decoder = #pair_decoder_t::decode_first(&mut variant)?;
+                let tag_decoder = #pair_decoder_t::first(&mut variant)?;
                 #decode_t::decode(tag_decoder)?
             };
 
@@ -629,49 +635,33 @@ impl<'a> Expander<'a> {
     fn encode_field_tag(
         &self,
         fields: &syn::Fields,
+        decls: &[TokenStream],
+        encode: TokenStream,
         tests: &[syn::Ident],
-    ) -> Option<(TokenStream, Option<TokenStream>)> {
+    ) -> TokenStream {
         let encoder_var = &self.tokens.encoder_var;
         let encoder_t = &self.tokens.encoder_t;
+        let pair_encoder_t = &self.tokens.pair_encoder_t;
 
-        match fields {
+        let setup = match fields {
             syn::Fields::Named(..) => {
                 let fields = calculate_tests(fields.len(), tests);
-
-                let setup = quote! {
-                    #encoder_t::encode_struct(#encoder_var, #fields)?
-                };
-
-                let pair_encoder_t = &self.tokens.pair_encoder_t;
-
-                let finish = Some(quote! {
-                    #pair_encoder_t::finish(#encoder_var)?;
-                });
-
-                Some((setup, finish))
+                quote!(#encoder_t::encode_struct(#encoder_var, #fields)?)
             }
             syn::Fields::Unnamed(..) => {
                 let fields = calculate_tests(fields.len(), tests);
-
-                let setup = quote! {
-                    #encoder_t::encode_tuple(#encoder_var, #fields)?
-                };
-
-                let pair_encoder_t = &self.tokens.pair_encoder_t;
-
-                let finish = Some(quote! {
-                    #pair_encoder_t::finish(#encoder_var)?;
-                });
-
-                Some((setup, finish))
+                quote!(#encoder_t::encode_tuple(#encoder_var, #fields)?)
             }
             syn::Fields::Unit => {
-                let setup = quote! {
-                    #encoder_t::encode_unit_struct(#encoder_var)?
-                };
-
-                Some((setup, None))
+                return quote!(#encoder_t::encode_unit_struct(#encoder_var));
             }
+        };
+
+        quote! {
+            #(#decls)*
+            let mut #encoder_var = #setup;
+            #encode
+            #pair_encoder_t::end(#encoder_var)
         }
     }
 
@@ -700,14 +690,7 @@ impl<'a> Expander<'a> {
                     self.encode_variant_fields(&variant.fields, needs, packing, default_field_tag)?;
 
                 // Special stuff needed to encode the field if its tagged.
-                let (setup, finish) = self.encode_field_tag(&variant.fields, &tests)?;
-
-                let encode = quote! {
-                    let mut #encoder_var = #setup;
-                    #encode
-                    #finish
-                };
-
+                let encode = self.encode_field_tag(&variant.fields, &[], encode, &tests);
                 (encode, patterns)
             }
             Packing::Packed => {
@@ -724,7 +707,7 @@ impl<'a> Expander<'a> {
             let Tokens {
                 encode_t,
                 encoder_t,
-                variant_encoder_t,
+                pair_encoder_t,
                 ..
             } = &self.tokens;
 
@@ -737,10 +720,13 @@ impl<'a> Expander<'a> {
 
             encode = quote! {{
                 let mut variant_encoder = #encoder_t::encode_variant(#encoder_var)?;
-                let tag_encoder = #variant_encoder_t::encode_variant_tag(&mut variant_encoder)?;
+                let tag_encoder = #pair_encoder_t::first(&mut variant_encoder)?;
                 #encode_t::encode(&#tag, tag_encoder)?;
-                let #encoder_var = #variant_encoder_t::encode_variant_value(variant_encoder)?;
-                #encode
+                let _ = {
+                    let #encoder_var = #pair_encoder_t::second(&mut variant_encoder)?;
+                    #encode
+                }?;
+                #pair_encoder_t::end(variant_encoder)
             }};
         }
 
@@ -816,7 +802,7 @@ impl<'a> Expander<'a> {
                     #(#decls)*
                     let mut pack = #encoder_t::encode_pack(#encoder_var)?;
                     #(#encoders)*
-                    #pack_encoder_t::finish(pack)?;
+                    #pack_encoder_t::end(pack)
                 }
             }
             _ => quote!(),
@@ -842,7 +828,7 @@ impl<'a> Expander<'a> {
 
         let body = match tagged {
             Packing::Tagged | Packing::Transparent => {
-                let tag = self.expand_field_tag(
+                let (tag, _) = self.expand_field_tag(
                     field_attr.rename.as_ref(),
                     default_field_tag,
                     index,
@@ -853,9 +839,9 @@ impl<'a> Expander<'a> {
 
                 quote_spanned! {
                     span => {
-                        let field_encoder = #pair_encoder_t::encode_first(&mut #encoder_var)?;
+                        let field_encoder = #pair_encoder_t::first(&mut #encoder_var)?;
                         #encode_t::encode(&#tag, field_encoder)?;
-                        let value_encoder = #pair_encoder_t::encode_second(&mut #encoder_var)?;
+                        let value_encoder = #pair_encoder_t::second(&mut #encoder_var)?;
                         #encode_path(#access, value_encoder)?;
                     }
                 }
@@ -921,26 +907,51 @@ impl<'a> Expander<'a> {
         let mut decls = Vec::with_capacity(fields_len);
         let mut patterns = Vec::with_capacity(fields_len);
         let mut assigns = Vec::with_capacity(fields_len);
+        let mut string_patterns = Vec::with_capacity(fields_len);
+        let mut output_variants = Vec::with_capacity(fields_len);
+
+        let tag_visitor_output = syn::Ident::new("TagVisitorOutput", self.input.ident.span());
+
+        let mut tag_methods = HashSet::new();
 
         for (index, field) in fields.enumerate() {
             let field_attr = attr::field_attrs(&self.cx, &field.attrs);
 
-            let tag =
+            let (tag, tag_method) =
                 self.expand_field_tag(field_attr.rename.as_ref(), default_field_tag, index, field)?;
 
+            let before = tag_methods.len();
+            tag_methods.insert(tag_method);
+
+            if before == 1 && tag_methods.len() > 1 {
+                self.cx
+                    .error_spanned_by(field, format!("#[{}({})] conflicting tag kind", ATTR, TAG));
+            }
+
             let (span, decode_path) = field_attr.decode_path(decode_t, field.span());
             let var = syn::Ident::new(&format!("v{}", index), span);
             decls.push(quote_spanned!(span => let mut #var = None;));
             let decode = quote_spanned!(span => #var = Some(#decode_path(#decoder_var)?));
 
-            let pair_decoder_t = &self.tokens.pair_decoder_t;
+            let output_tag = match tag_method {
+                TagMethod::String => {
+                    let variant = syn::Ident::new(&format!("Variant{}", index), field.span());
+                    let mut path = syn::Path::from(tag_visitor_output.clone());
+                    path.segments.push(syn::PathSegment::from(variant.clone()));
 
-            patterns.push(quote! {
-                #tag => {
-                    let #decoder_var = #pair_decoder_t::decode_second(#decoder_var)?;
-                    #decode;
+                    string_patterns.push(quote!(#tag => Ok(#path)));
+                    output_variants.push(variant.clone());
+
+                    syn::Expr::Path(syn::ExprPath {
+                        attrs: Vec::new(),
+                        qself: None,
+                        path,
+                    })
                 }
-            });
+                TagMethod::Default => tag.clone(),
+            };
+
+            patterns.push((output_tag, decode));
 
             let field = match &field.ident {
                 Some(ident) => quote!(#ident),
@@ -971,24 +982,101 @@ impl<'a> Expander<'a> {
             },
         };
 
-        let decode_next = match field_kind {
+        let next = match field_kind {
             FieldKind::Struct => {
                 let struct_decoder_t = &self.tokens.struct_decoder_t;
-                quote!(#struct_decoder_t::decode_field)
+                quote!(#struct_decoder_t::next)
             }
             FieldKind::Tuple => {
                 let struct_decoder_t = &self.tokens.struct_decoder_t;
-                quote!(#struct_decoder_t::decode_field)
+                quote!(#struct_decoder_t::next)
             }
         };
 
+        let tag_method = tag_methods.into_iter().next().unwrap_or_default();
+
         let pair_decoder_t = &self.tokens.pair_decoder_t;
         let decode_t = &self.tokens.decode_t;
 
-        let decode_tag = quote! {{
-            let index_decoder = #pair_decoder_t::decode_first(&mut #decoder_var)?;
-            #decode_t::decode(index_decoder)?
-        }};
+        let (decode_tag, unsupported_pattern, patterns, output_enum) = match tag_method {
+            TagMethod::String => {
+                let reference_visitor_t = &self.tokens.reference_visitor_t;
+                let phantom_data = &self.tokens.phantom_data;
+                let fmt = &self.tokens.fmt;
+                let error_t = &self.tokens.error_t;
+
+                // Declare a tag visitor, allowing string tags to be decoded by
+                // decoders that owns the string.
+                let decode_tag = quote! {{
+                    struct TagVisitor<E>(#phantom_data<E>);
+
+                    impl<'de, E> #reference_visitor_t<'de> for TagVisitor<E> where E: #error_t {
+                        type Target = str;
+                        type Ok = Result<#tag_visitor_output, Box<str>>;
+                        type Error = E;
+
+                        #[inline]
+                        fn expecting(&self, f: &mut #fmt::Formatter<'_>) -> #fmt::Result {
+                            write!(f, "string tag")
+                        }
+
+                        #[inline]
+                        fn visit_borrowed(self, string: &'de Self::Target) -> Result<Self::Ok, Self::Error> {
+                            self.visit_any(string)
+                        }
+
+                        #[inline]
+                        fn visit_any(self, string: &Self::Target) -> Result<Self::Ok, Self::Error> {
+                            Ok(match string {
+                                #(#string_patterns,)*
+                                _ => Err(string.into()),
+                            })
+                        }
+                    }
+
+                    let index_decoder = #pair_decoder_t::first(&mut #decoder_var)?;
+                    #decoder_t::decode_string(index_decoder, TagVisitor::<D::Error>(#phantom_data))?
+                }};
+
+                let patterns = patterns
+                    .into_iter()
+                    .map(|(tag, decode)| (quote!(Ok(#tag)), decode))
+                    .collect::<Vec<_>>();
+
+                let output_enum = quote! {
+                    enum #tag_visitor_output {
+                        #(#output_variants,)*
+                    }
+                };
+
+                (decode_tag, quote!(Err(tag)), patterns, Some(output_enum))
+            }
+            TagMethod::Default => {
+                let decode_tag = quote! {{
+                    let index_decoder = #pair_decoder_t::first(&mut #decoder_var)?;
+                    #decode_t::decode(index_decoder)?
+                }};
+
+                let patterns = patterns
+                    .into_iter()
+                    .map(|(tag, decode)| (quote!(#tag), decode))
+                    .collect::<Vec<_>>();
+
+                (decode_tag, quote!(tag), patterns, None)
+            }
+        };
+
+        let patterns = patterns
+            .into_iter()
+            .map(|(tag, decode)| {
+                quote! {
+                    #tag => {
+                        let #decoder_var = #pair_decoder_t::second(#decoder_var)?;
+                        #decode;
+                    }
+                }
+            })
+            .collect::<Vec<_>>();
 
         let skip_field = quote! {
             #pair_decoder_t::skip_second(#decoder_var)?
@@ -996,10 +1084,10 @@ impl<'a> Expander<'a> {
 
         let unsupported = match variant_tag {
             Some(variant_tag) => quote! {
-                <D::Error as #error_t>::unsupported_variant_field(#type_name, #variant_tag, tag)
+                <D::Error as #error_t>::invalid_variant_field_tag(#type_name, #variant_tag, tag)
             },
             None => quote! {
-                <D::Error as #error_t>::unsupported_field(#type_name, tag)
+                <D::Error as #error_t>::invalid_field_tag(#type_name, tag)
             },
         };
 
@@ -1015,7 +1103,7 @@ impl<'a> Expander<'a> {
             quote! {
                 match tag {
                     #(#patterns,)*
-                    tag => {
+                    #unsupported_pattern => {
                         if !#skip_field {
                             return Err(#unsupported);
                         }
@@ -1026,10 +1114,11 @@ impl<'a> Expander<'a> {
 
         Some(quote! {
             #(#decls;)*
+            #output_enum
 
             let mut #decoder_var = #declare;
 
-            while let Some(mut #decoder_var) = #decode_next(&mut #decoder_var)? {
+            while let Some(mut #decoder_var) = #next(&mut #decoder_var)? {
                 let tag #tag_type = #decode_tag;
                 #body
             }
@@ -1093,7 +1182,7 @@ impl<'a> Expander<'a> {
             Some(quote! {{
                 let mut unpack = #decoder_t::decode_pack(#decoder_var)?;
                 let this = #path { #(#assign),* };
-                #pack_decoder_t::finish(unpack)?;
+                #pack_decoder_t::end(unpack)?;
                 this
             }})
         }
@@ -1126,10 +1215,14 @@ impl<'a> Expander<'a> {
         default_field_tag: DefaultTag,
         index: usize,
         field: &syn::Field,
-    ) -> Option<syn::Expr> {
-        let lit = match (rename, default_field_tag, &field.ident) {
-            (Some((_, rename)), _, _) => return Some(self.rename_lit(rename)),
-            (None, DefaultTag::Index, _) => usize_int(index, field.span()).into(),
+    ) -> Option<(syn::Expr, TagMethod)> {
+        let (lit, tag_method) = match (rename, default_field_tag, &field.ident) {
+            (Some((_, rename)), _, _) => {
+                return Some((self.rename_lit(rename), self.determine_tag_method(rename)))
+            }
+            (None, DefaultTag::Index, _) => {
+                (usize_int(index, field.span()).into(), TagMethod::Default)
+            }
             (None, DefaultTag::Name, None) => {
                 self.cx.error_spanned_by(
                     field,
@@ -1140,15 +1233,18 @@ impl<'a> Expander<'a> {
                 );
                 return None;
             }
-            (None, DefaultTag::Name, Some(ident)) => {
-                syn::LitStr::new(&ident.to_string(), ident.span()).into()
-            }
+            (None, DefaultTag::Name, Some(ident)) => (
+                syn::LitStr::new(&ident.to_string(), ident.span()).into(),
+                TagMethod::String,
+            ),
         };
 
-        Some(syn::Expr::Lit(syn::ExprLit {
+        let tag = syn::Expr::Lit(syn::ExprLit {
             attrs: Vec::new(),
             lit,
-        }))
+        });
+
+        Some((tag, tag_method))
     }
 
     /// Process rename literal to ensure it's always typed.
@@ -1164,6 +1260,22 @@ impl<'a> Expander<'a> {
             expr => expr.clone(),
         }
     }
+
+    /// Try and determine tag method from the given expression.
+    fn determine_tag_method(&self, expr: &syn::Expr) -> TagMethod {
+        let lit = match expr {
+            syn::Expr::Lit(lit) => lit,
+            _ => return TagMethod::Default,
+        };
+
+        match lit {
+            syn::ExprLit {
+                lit: syn::Lit::Str(..),
+                ..
+            } => TagMethod::String,
+            _ => TagMethod::Default,
+        }
+    }
 }
 
 fn calculate_tests(count: usize, tests: &[syn::Ident]) -> TokenStream {
